/****************************************************************************

                      (C) copyright Coding Technologies (2002)
                               All Rights Reserved 

 This software module was developed by Coding Technologies (CT). This is
 company confidential information and the property of CT, and can not be
 reproduced or disclosed in any form without written authorization of CT.

 Those intending to use this software module for other purposes are advised
 that this infringe existing or pending patents. CT has no liability for
 use of this software module or derivatives thereof in any implementation.
 Copyright is not released for any means.  CT retains full right to use the
 code for its own purpose, assign or sell the code to a third party and to
 inhibit any user or third party from using the code. This copyright notice
 must be included in all copies or derivative works.

 $Id: fract_test.cpp,v 1.2 2003/08/13 13:34:01 ehr Exp $

****************************************************************************/
/*!
  \file   fract_test.cpp
  \brief  Testfile for fractional library $Revision: 1.2 $
*/

#include "fract.h"
#include <stdio.h>
#include <math.h>
#include <stdlib.h>

#define NUMBER_OPERATION_TESTS 100000

#define FRACT     SFract
#define FRACTBITS SFRACT_BITS

/***************************************************************************/
/*!

  \brief  Tiny class for operations test

  This class is used to test the operations of the fractional library

****************************************************************************/

class operation_test {
public:
  operation_test();
  void show_statistics();
  void add(int i);
  void test(long double a, long double b);
protected:
  virtual int operation(long double a, long double b, 
                        FRACT sa, FRACT sb, long double &c, FRACT &sc) = 0;
  long bit_diff_max;
  char print_oper[3];
private:
  long cnt[FRACTBITS];
  long total;
};

operation_test::operation_test() {
  int i;

  for (i = 0; i < FRACTBITS; i++) {
    cnt[i] = 0;
  }
  total = 0;
  bit_diff_max = 1;
}

void
operation_test::show_statistics() {
  int i;

  for (i = 0; i < FRACTBITS; i++) {
    printf("%2d bit: %10ld (%02.2f %%) \n", 
           i, cnt[i], (((float)cnt[i])/((float)total)*100.0));
  }
  printf("\n");
}

void
operation_test::add(int i) {
  cnt[i]++;
  total++;
}

void
operation_test::test(long double a, long double b) {
  long double c;
  double diff, min_val, level_diff, bit_diff;
  FRACT fa, fb, fc;
  long lc, lfc;

  fa = (double)a;
  fb = (double)b;
  a = (double)fa;
  b = (double)fb;

  min_val = pow(2, -(FRACTBITS-1));

  if(operation(a, b, fa, fb, c, fc) != 0) return;

  lc = (long)((FRACT)(double)c);
  lfc = (long)fc;

  diff = (double)fc - (double)c;
  /* level_diff = fabs(diff)/fabs(min_val); */
  level_diff = fabs(lc - lfc);
  bit_diff = ceil(log(level_diff+1.0)/log(2.0));
  add((int)bit_diff);

  if (bit_diff > bit_diff_max) {
    printf("----------------------------------------------------------\n");
    if (c < 1.0)
      printf("       %.10Lf %s %.10Lf = %.10Lf (%10ld)\n", 
             a, print_oper, b, c, lc);
    else
      printf("       %.10Lf %s %.10Lf = %.10Lf (%10ld saturated)\n", 
             a, print_oper, b, c, lc);
    printf("       %.10f %s %.10f = %.10f (%10ld)\n\n", 
           (double)fa, print_oper, (double)fb, (double)fc, lfc);
    printf("diff = %.10f (min = %.10f) ==> difference of %.2f level  \n\n", 
           diff, min_val, level_diff);
    if (c > 1.0)
      printf("       --> level diff. calculation using saturated floating-point value\n");
  }
}

/***************************************************************************/
/*!

  \brief  test the fractional multiplication

  Test the fractional multiplication

****************************************************************************/

class test_multiplication : public operation_test {
public:
  test_multiplication() { char tmp_print_op[3] = "* ";  print_oper = tmp_print_op; }
protected:
  int operation(long double a, long double b, 
                FRACT sa, FRACT sb, long double &c, FRACT &sc);
};

int
test_multiplication::operation(long double a, long double b, 
                               FRACT sa, FRACT sb, long double &c, FRACT &sc) {
  c = a * b;
  sc = (FRACT)(sa * sb);
  return 0;
}

/***************************************************************************/
/*!

  \brief  test the fractional division

  Test the fractional division

****************************************************************************/

class test_division : public operation_test {
public:
  test_division() { char tmp_print_op[3] = "/ ";  print_oper = tmp_print_op; }
protected:
  int operation(long double a, long double b, 
                FRACT sa, FRACT sb, long double &c, FRACT &sc);
};

int
test_division::operation(long double a, long double b, 
                         FRACT sa, FRACT sb, long double &c, FRACT &sc) {
  long double min_val = pow(2, -(FRACTBITS-1));
  if( (b == 0.0) || (fabs(b) < fabs(a)) || (fabs(a-b) <  min_val) ) return -1;

  c = a / b;
  sc = (FRACT)(sa / sb);
  return 0;
}

/***************************************************************************/
/*!

  \brief  test the fractional addition

  Test the fractional addition

****************************************************************************/

class test_addition : public operation_test {
public:
  test_addition() { char tmp_print_op[3] = "+ ";  print_oper = tmp_print_op; }
  int operation(long double a, long double b, 
                FRACT sa, FRACT sb, long double &c, FRACT &sc);
};

int
test_addition::operation(long double a, long double b, 
                         FRACT sa, FRACT sb, long double &c, FRACT &sc) {
  c = a + b;
  sc = (FRACT)(sa + sb);
  return 0;
}

/***************************************************************************/
/*!

  \brief  test the fractional subtraction

  Test the fractional subtraction

****************************************************************************/

class test_subtraction : public operation_test {
public:
  test_subtraction() { char tmp_print_op[3] = "- ";  print_oper = tmp_print_op; }
  int operation(long double a, long double b, FRACT sa, 
FRACT sb, long double &c, FRACT &sc);
};

int
test_subtraction::operation(long double a, long double b, 
                            FRACT sa, FRACT sb, long double &c, FRACT &sc) {
  c = a - b;
  sc = (FRACT)(sa - sb);
  return 0;
}

/***************************************************************************/
/*!

  \brief  test of fractional precision

  Test the precision of the fractional classes

****************************************************************************/

void test_print_precision(long double d) {
  fract  f  = ((double)d);
  dfract df = ((double)d);

  printf("--> double = %.16Lf \n\n", d);
  printf("    fract  = %.16f (%ld)\n", (double) f, (long)f);
  printf("    dfract = %.16f (%lld)\n", (double) df, (long long)df);
  printf("\n");
}

/***************************************************************************/
/*!

  \brief  test of fractional addition

  Test the addition operation of the fractional classes

****************************************************************************/

void test_print_addition(double d1, double d2) {
  double d;
  dfract df;
  fract  f1, f2;
  f1 = d1; d1 = f1;
  f2 = d2; d2 = f2;

  df = f1 + f2;
  d = d1 + d2;

  printf(" double: %.10f + (%.10f) = %.10f (%10ld)\n", 
         d1, d2, d, (long)((fract)d));
  printf(" fract:  %.10f + (%.10f) = %.10f (%10ld)\n", 
         (double)f1, (double)f2, (double)df, (long)df);
  printf("\n");
}

/***************************************************************************/
/*!

  \brief  test of fractional subtraction

  Test the subtraction operation of the fractional classes

****************************************************************************/

void test_print_subtraction(double d1, double d2) {
  double d;
  dfract df;
  fract  f1, f2;
  f1 = d1; d1 = f1;
  f2 = d2; d2 = f2;

  df = f1 - f2;
  d = d1 - d2;

  printf(" double: %.10f - (%.10f) = %.10f (%10ld)\n", 
         d1, d2, d, (long)((fract)d));
  printf(" fract:  %.10f - (%.10f) = %.10f (%10ld)\n", 
         (double)f1, (double)f2, (double)df, (long)df);
  printf("\n");
}

/***************************************************************************/
/*!

  \brief  test of fractional multiplication

  Test the multiplication operation of the fractional classes

****************************************************************************/

void test_print_multiplication(double d1, double d2) {
  double d;
  dfract df;
  fract  f1, f2;
  f1 = d1; d1 = f1;
  f2 = d2; d2 = f2;

  df = f1 * f2;
  d = d1 * d2;

  printf(" double: %.10f * (%.10f) = %.10f (%10ld)\n", 
         d1, d2, d, (long)((fract)d));
  printf(" fract:  %.10f * (%.10f) = %.10f (%10ld)\n", 
         (double)f1, (double)f2, (double)df, (long)df);
  printf("\n");
}

/***************************************************************************/
/*!

  \brief  test of fractional multiplication

  Test the multiplication operation of the fractional classes

****************************************************************************/

void test_print_multiplication(double d1, int i2) {
  double d;
  dfract df;
  fract  f1;
  f1 = d1; d1 = f1;

  df = f1 * i2;
  d = d1 * i2;

  printf(" double: %.10f * (%10d) = %.10f (%10ld)\n", 
         d1, i2, d, (long)((fract)d));
  printf(" fract:  %.10f * (%10d) = %.10f (%10ld)\n", 
         (double)f1, i2, (double)df, (long)df);
  printf("\n");
}


/***************************************************************************/
/*!

  \brief  test of fractional division

  Test the division operation of the fractional classes

****************************************************************************/

void test_print_division(double d1, double d2) {
  double d;
  dfract df;
  fract  f1, f2;
  f1 = d1; d1 = f1;
  f2 = d2; d2 = f2;

  df = f1 / f2;
  d = d1 / d2;

  printf(" double: %.10f / (%.10f) = %.10f (%10ld)\n", 
         d1, d2, d, (long)((fract)d));
  printf(" fract:  %.10f / (%.10f) = %.10f (%10ld)\n", 
         (double)f1, (double)f2, (double)df, (long)df);
  printf("\n");
}

/***************************************************************************/
/*!

  \brief  test of fractional division

  Test the division operation of the fractional classes

****************************************************************************/

void test_print_division(double d1, int i2) {
  double d;
  dfract df, df1;
  df1 = d1; d1 = df1;

  df = df1 / i2;
  d = d1 / i2;

  printf(" double: %.10f / (%10d) = %.10f (%10ld)\n", 
         d1, i2, d, (long)((fract)d));
  printf(" dfract: %.10f / (%10d) = %.10f (%10ld)\n", 
         (double)df1, i2, (double)df, (long)df);
  printf("\n");
}

/***************************************************************************/
/*!

  \brief  Main function

  Main function of the fractional library testing application

****************************************************************************/

int main() {

  {
    printf("==========================================================\n");
    printf("==       Testing of precision and extreme values        ==\n");
    printf("==========================================================\n");

    test_print_precision( 0.11111111111111111);
    test_print_precision( 0.99999999999999999);
    test_print_precision(-0.99999999999999999);
    test_print_precision(pow(2, -(SFRACT_BITS-1)));
    test_print_precision(pow(2, -(FRACT_BITS-1)));
    test_print_precision(pow(2, -(DFRACT_BITS-1)));
  }

  {
    printf("==========================================================\n");
    printf("==       Testing of operations on  \"simple\" values      ==\n");
    printf("==========================================================\n");

    test_print_addition(0.1, 0.2);
    test_print_subtraction(0.2, 0.1);
    test_print_multiplication(0.5, 0.4);
    test_print_multiplication(0.2, 3);
    test_print_division(0.4, 0.5);
    test_print_division(0.6, 3);

    printf("==========================================================\n");
    printf("==       Testing of operations on extreme values        ==\n");
    printf("==========================================================\n");

    test_print_addition(1.0, 1.0);
    test_print_addition(-1.0, 1.0);
    test_print_addition(-1.0, -1.0);
    test_print_addition(1.0, 0);
    test_print_addition(pow(2, -(FRACTBITS-1)), pow(2, -(FRACTBITS-1)));
    test_print_addition(-pow(2, -(FRACTBITS-1)), -pow(2, -(FRACTBITS-1)));
    test_print_addition(-pow(2, -(FRACTBITS-2)), pow(2, -(FRACTBITS-1)));
    test_print_addition(1.0, pow(2, -(FRACTBITS-1)));

    test_print_subtraction(1.0, 1.0);
    test_print_subtraction(1.0, -1.0);
    test_print_subtraction(-1.0, 1.0);
    test_print_subtraction(-1.0, pow(2, -(FRACTBITS-1)));
    test_print_subtraction(pow(2, -(FRACTBITS-2)), pow(2, -(FRACTBITS-1)));
    test_print_subtraction(pow(2, -(FRACTBITS-1)), pow(2, -(FRACTBITS-2)));

    test_print_multiplication(1.0, 1.0);
    test_print_multiplication(1.0, 2);
    test_print_multiplication(-1.0, 10);
    test_print_multiplication(1.0, int(pow(2, OVRFLW_BITS+1)));
    test_print_multiplication(-1.0, int(pow(2, OVRFLW_BITS+1)));
    test_print_multiplication(-1.0, -1);
    test_print_multiplication(-0.0, 0.0);
    test_print_multiplication(-0.0, (int)0);
    test_print_multiplication(pow(2, -(FRACTBITS/2)), pow(2, -(FRACTBITS/2)));
    test_print_multiplication(pow(2, -(FRACTBITS/2-1)), pow(2, -(FRACTBITS/2-1)));
    test_print_multiplication(pow(2, -(FRACTBITS/2-1)), -pow(2, -(FRACTBITS/2-1)));
    test_print_multiplication(-pow(2, -(FRACTBITS/2-1)), -pow(2, -(FRACTBITS/2-1)));
    test_print_multiplication(pow(2, -(FRACTBITS-1)), pow(2, -(FRACTBITS-1)));
    test_print_multiplication(pow(2, -(FRACTBITS-1)), (int)2);


    test_print_division(pow(2, -(FRACTBITS-1)), 1.0);
    test_print_division(pow(2, -(FRACTBITS-1)), -pow(2, -(FRACTBITS-2)));
    test_print_division(0.0, pow(2, -(FRACTBITS-1)));
    test_print_division(0.5-pow(2, -(FRACTBITS-1)), 0.5);
    test_print_division(1.0, -10);
    test_print_division(pow(2, -(FRACTBITS-2)), 2);
    test_print_division(pow(2, -(FRACTBITS-1)), -2);
    test_print_division(1.0, (int)pow(2, 6));
    test_print_division(1.0, (int)pow(2, (FRACTBITS-3)));
    test_print_division(1.0, (int)pow(2, (FRACTBITS-2)));
    test_print_division(1.0, -(int)pow(2, (FRACTBITS-2)));
  }

  {
    long i, j;
    operation_test *op_tst[4];
    op_tst[0] = new test_addition;
    op_tst[1] = new test_subtraction;
    op_tst[2] = new test_multiplication;
    op_tst[3] = new test_division;

    printf("==========================================================\n");
    printf("==                Testing of operators                  ==\n");
    printf("==========================================================\n");

    for (i = 0; i < 4; i++) {
      for (j = 0; j < NUMBER_OPERATION_TESTS; j++) {
        long double a = rand()/(RAND_MAX+1.0);
        long double b = rand()/(RAND_MAX+1.0);
        op_tst[i]->test(a, b);
      }
    }

    printf("Statistics for addition\n");    
    op_tst[0]->show_statistics();
    printf("Statistics for subtraction\n");    
    op_tst[1]->show_statistics();
    printf("Statistics for multiplication\n");
    op_tst[2]->show_statistics();
    printf("Statistics for division\n");    
    op_tst[3]->show_statistics();
  }

  {
    FRACT sf1, sf2;
    int j;
    int success = 1;

    printf("==========================================================\n");
    printf("==           Testing of comparison operators            ==\n");
    printf("==========================================================\n");

    for (j = 0; j < NUMBER_OPERATION_TESTS; j++) {
      double a = rand()/(RAND_MAX+1.0);
      double b = rand()/(RAND_MAX+1.0);

      if (a - b > pow(2, -(FRACTBITS-1))) {
        if (a < b) {
          sf1 = a;
          sf2 = b;
          if (!(sf1 < sf2)) {
            printf("<  :  Failure for %f <  %f (%ld <  %ld)\n", 
                   a, b, (long)sf1, (long)sf2);
            success = 0;
          }
          if (!(sf1 <= sf2)) {
            printf("<= :  Failure for %f <= %f (%ld <= %ld)\n", 
                   a, b, (long)sf1, (long)sf2);
            success = 0;
          }
        }
        else {
          sf1 = a;
          sf2 = b;
          if (!(sf1 > sf2)) {
            printf(">  :  Failure for %f >  %f (%ld >  %ld)\n", 
                   a, b, (long)sf1, (long)sf2);
            success = 0;
          }
          if (!(sf1 >= sf2)) {
            printf(">= :  Failure for %f >= %f (%ld >= %ld)\n", 
                   a, b, (long)sf1, (long)sf2);
            success = 0;
          }
        }
      }
    }
    for (j = 0; j < NUMBER_OPERATION_TESTS; j++) {
      double a = rand()/(RAND_MAX+1.0);
   
      sf1 = a;
      sf2 = a;
      if (!(sf1 == sf2)) {
        printf("== :  Failure for %f == %f (%ld == %ld)\n", 
               a, a, (long)sf1, (long)sf2);
        success = 0;
      }
      if (!(sf1 >= sf2)) {
        printf(">= :  Failure for %f >= %f (%ld >= %ld)\n", 
               a, a, (long)sf1, (long)sf2);
        success = 0;
      }
      if (!(sf1 <= sf2)) {
        printf("<= :  Failure for %f <= %f (%ld <= %ld)\n", 
               a, a, (long)sf1, (long)sf2);
        success = 0;
      }   
      sf2 = a - pow(2, -(FRACTBITS-2));
      if (!(sf1 != sf2)) {
        printf("!= :  Failure for %f != %f (%ld != %ld)\n", 
               a, a, (long)sf1, (long)sf2);
        success = 0;
      }
    }
    if (success) {
      printf("Testing of comparison operators successfull\n\n");
    } else {
      printf("Testing of comparison operators failed\n\n");
    }
  }

  {
    double d1, d2;
    FRACT sf1, sf2;
    int i, j;
    int success = 1;

    printf("==========================================================\n");
    printf("==           Testing of shifting operators              ==\n");
    printf("==========================================================\n");

    for (j = 1; j < FRACTBITS; j++) {
      for (i = 0; i < NUMBER_OPERATION_TESTS; i++) {
        d1 = rand()/(RAND_MAX+1.0);

        sf1 = d1;
        d1 = (double) sf1;

        d2 = d1 * pow(2, -j);
        sf2 = sf1 >> j;
        if (!((long)(FRACT)d2 == (long)sf2)) {
          printf("Failure for >> %d\n", j);
          printf(" double: %f >> %d = %f (%ld >> %d = %ld)\n", 
                 d1, j, d2, (long)(FRACT)d1, j, (long)(FRACT)d2);
          printf(" frac:   %f >> %d = %f (%ld >> %d = %ld)\n\n", 
                 (double)sf1, j, (double)sf2, (long)sf1, j, (long)sf2);
          success = 0;
        }
      }
    }
    for (j = 1; j < 15; j++) {
      for (i = 0; i < NUMBER_OPERATION_TESTS; i++) {
        d1 = rand()/(RAND_MAX+1.0);

        sf1 = d1;
        d1 = (double) sf1;

        d2 = d1 * pow(2, j);
        sf2 = sf1 << j;
        if (!((long)(FRACT)d2 == (long)sf2)) {
          printf("Failure for << %d\n", j);
          printf(" double: %f << %d = %f (%ld << %d = %ld)\n", 
                 d1, j, d2, (long)(FRACT)d1, j, (long)(FRACT)d2);
          printf(" frac:   %f << %d = %f (%ld << %d = %ld)\n\n", 
                 (double)sf1, j, (double)sf2, (long)sf1, j, (long)sf2);
          success = 0;
        }
      }
    }
       
    if (success) {
      printf("Testing of shifting operators successfull\n\n");
    } else {
      printf("Testing of shifting operators failed\n\n");
    }
  }

} 
